from os import PathLike
from typing import Literal, Union


class DeferredFileWriter:
    """A class that provides a context manager for deferring writing or appending to a
    file until a write method is called.

    Attributes
    ----------
    filename
        The file path to open
    mode
        The mode to use.

    Examples
    --------
    >>> with DeferredFileWrite("new_file.txt", "w") as f:
    ...     # file is not yet created
    ...     f.write("contents")
    ...     # file is now created, but not closed
    ... # file is now closed

    """

    def __init__(self, filename: Union[str, PathLike], mode: Literal["w", "a"], **kwargs):
        self.filename = filename
        self.mode = mode
        self.file = None
        self.kwargs = kwargs
        self._is_open = False

    def _ensure_open(self):
        if not self._is_open:
            self.file = open(self.filename, self.mode, **self.kwargs)
            self._is_open = True

    def write(self, data):
        self._ensure_open()
        self.file.write(data)

    def writelines(self, lines):
        self._ensure_open()
        self.file.writelines(lines)

    def close(self):
        if self._is_open:
            self.file.close()
            self._is_open = False

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.close()



"""
Configuration file adapter.
"""

from __future__ import annotations

import copy
import fnmatch
import io
import logging
import os
from pathlib import Path
import random
import string
from typing import cast, TYPE_CHECKING

from ruamel.yaml import YAML

from hpcflow.sdk.core.validation import Schema, get_schema

from hpcflow.sdk.config.errors import (
    ConfigChangeFileUpdateError,
    ConfigDefaultValidationError,
    ConfigFileInvocationIncompatibleError,
    ConfigFileInvocationUnknownMatchKey,
    ConfigFileValidationError,
    ConfigInvocationKeyNotFoundError,
    ConfigValidationError,
)

if TYPE_CHECKING:
    from typing import Any
    from ..typing import PathLike
    from .config import Config, ConfigOptions
    from .types import ConfigDict, DefaultConfiguration, InvocationDescriptor


class ConfigFile:
    """
    Configuration file.

    Parameters
    ----------
    directory:
        The directory containing the configuration file.
    logger:
        Where to log messages.
    config_options:
        Configuration options.
    """

    def __init__(self, directory, logger: logging.Logger, config_options: ConfigOptions):
        #: Where to log messages.
        self.logger = logger
        #: The directory containing the configuration file.
        self.directory = self._resolve_config_dir(
            config_opt=config_options,
            logger=self.logger,
            directory=directory,
        )

        self._configs: list[Config] = []

        # set by _load_file_data:
        self.__path: Path | None = None
        self.__contents: str | None = None
        self.__data: ConfigDict | None = None
        self.__data_rt: ConfigDict | None = None

        self._load_file_data(config_options)
        self.file_schema = self._validate(self.__data)

    @property
    def data(self) -> ConfigDict:
        """
        The parsed contents of the config file.
        """
        d = self.__data
        assert d is not None
        return d

    @property
    def data_rt(self) -> ConfigDict:
        """
        The parsed contents of the config file where the alternate parser was used.
        """
        drt = self.__data_rt
        assert drt is not None
        return drt

    @property
    def path(self) -> Path:
        """
        The path to the config file.
        """
        p = self.__path
        assert p is not None
        return p

    @property
    def contents(self) -> str:
        """
        The cached contents of the config file.
        """
        c = self.__contents
        assert c is not None
        return c

    @staticmethod
    def select_invocation(
        configs: dict[str, Any],
        run_time_info: dict[str, Any],
        path: PathLike,
        config_key: str | None = None,
    ) -> str:
        """Select a matching configuration for this invocation using run-time info."""
        if not config_key:
            all_matches = {}  # keys are config keys; values are lengths of match dict
            for c_name_i, c_dat_i in configs.items():
                # for a config to "match", each "match key" must match the relevant run
                # time info attribute. If a "match key" has multiple values, at least
                # one value must match the run time info attribute:
                for match_k, match_v in c_dat_i["invocation"]["match"].items():
                    # test for a matching glob pattern (where multiple may be specified):
                    if not isinstance(match_v, list):
                        match_v = [match_v]

                    try:
                        k_value = run_time_info[match_k]
                    except KeyError:
                        raise ConfigFileInvocationUnknownMatchKey(match_k)

                    if not any(
                        fnmatch.filter(names=[k_value], pat=match_i)
                        for match_i in match_v
                    ):
                        break
                else:
                    all_matches[c_name_i] = len(c_dat_i["invocation"]["match"])

            if not all_matches:
                raise ConfigFileInvocationIncompatibleError(config_key)
            # for multiple matches select the more specific one:
            config_key = max(all_matches.items(), key=lambda x: x[1])[0]

        elif config_key not in configs:
            raise ConfigInvocationKeyNotFoundError(config_key, path, list(configs))

        return config_key

    def _validate(self, data: dict[str, Any] | None) -> Schema:
        file_schema = get_schema("config_file_schema.yaml")
        if not (file_validated := file_schema.validate(data)).is_valid:
            raise ConfigFileValidationError(file_validated.get_failures_string())
        return file_schema

    def get_invoc_data(self, config_key: str) -> DefaultConfiguration:
        """
        Get the invocation data for the given configuration.

        Parameters
        ----------
        config_key: str
            The name of the configuration within the configuration file.
        """
        return self.data["configs"][config_key]

    def get_invocation(self, config_key: str) -> InvocationDescriptor:
        """
        Get the invocation for the given configuration.

        Parameters
        ----------
        config_key: str
            The name of the configuration within the configuration file.
        """
        return self.get_invoc_data(config_key)["invocation"]

    def save(self) -> None:
        """
        Write the (modified) configuration to the configuration file.
        """
        new_data = copy.deepcopy(self.data)
        new_data_rt = copy.deepcopy(self.data_rt)
        new_contents = ""

        modified_names: list[str] = []
        for config in self._configs:
            modified_names.extend(config._modified_keys)
            modified_names.extend(config._unset_keys)

            new_data_config = new_data["configs"][config._config_key]["config"]
            new_data_rt_config = new_data_rt["configs"][config._config_key]["config"]
            new_data_config.update(config._modified_keys)
            new_data_rt_config.update(config._modified_keys)

            for k in config._unset_keys:
                del cast("dict", new_data_config)[k]
                del cast("dict", new_data_rt_config)[k]

        try:
            new_contents = self._dump(new_data_rt)
        except Exception as err:
            raise ConfigChangeFileUpdateError(names=modified_names, err=err) from None

        self.__data = new_data
        self.__data_rt = new_data_rt
        self.__contents = new_contents

        for config in self._configs:
            config._unset_keys = set()
            config._modified_keys = {}

    @staticmethod
    def _resolve_config_dir(
        config_opt: ConfigOptions,
        logger: logging.Logger,
        directory: str | Path | None = None,
    ) -> Path:
        """Find the directory in which to locate the configuration file.

        If no configuration directory is specified, look first for an environment variable
        (given by config option `directory_env_var`), and then in the default
        configuration directory (given by config option `default_directory`).

        The configuration directory will be created if it does not exist.

        Parameters
        ----------
        directory
            Directory in which to find the configuration file. Optional.

        Returns
        -------
        directory : Path
            Absolute path to the configuration directory.

        """

        if not directory:
            path = Path(
                os.getenv(config_opt.directory_env_var, config_opt.default_directory)
            ).expanduser()
        else:
            path = Path(directory)

        if not path.is_dir():
            logger.debug(
                f"Configuration directory does not exist. Generating here: {str(path)!r}."
            )
            path.mkdir()
        else:
            logger.debug(f"Using configuration directory: {str(path)!r}.")

        return path.resolve()

    def _dump(self, config_data: ConfigDict, path: Path | None = None) -> str:
        """Dump the specified config data to the specified config file path.

        Parameters
        ----------
        config_data
            New configuration file data that will be dumped using the "round-trip" dumper.
        path
            Path to dump the config file data to. If not specified the `path` instance
            attribute will be used. If the file already exists, an "atomic-ish" overwrite
            will be used, where we firstly create a temporary file, which then replaces
            the existing file.

        Returns
        -------
        new_contents
            String contents of the new file.

        """

        if path is None:
            path = self.path

        yaml = YAML(typ="rt")
        if path.exists():
            # write a new temporary config file
            cfg_tmp_file = path.with_suffix(path.suffix + ".tmp")
            self.logger.debug(f"Creating temporary config file: {cfg_tmp_file!r}.")
            with cfg_tmp_file.open("wt", newline="\n") as fh:
                yaml.dump(config_data, fh)

            # atomic rename, overwriting original:
            self.logger.debug("Replacing original config file with temporary file.")
            os.replace(src=cfg_tmp_file, dst=path)

        else:
            with path.open("w", newline="\n") as handle:
                yaml.dump(config_data, handle)

        buff = io.BytesIO()
        yaml.dump(config_data, buff)
        new_contents = str(buff.getvalue())

        return new_contents

    def add_default_config(
        self, config_options: ConfigOptions, name: str | None = None
    ) -> str:
        """Add a new default config to the config file, and create the file if it doesn't
        exist."""

        if self.path.exists():
            is_new_file = False
            new_data: ConfigDict = copy.deepcopy(self.data)
            new_data_rt: ConfigDict = copy.deepcopy(self.data_rt)
        else:
            is_new_file = True
            new_data = {"configs": {}}
            new_data_rt = {"configs": {}}

        if not name:
            name = "".join(random.choices(string.ascii_letters, k=6))

        def_config = copy.deepcopy(config_options.default_config)
        new_config = {name: def_config}

        new_data["configs"].update(new_config)
        new_data_rt["configs"].update(new_config)

        try:
            if is_new_file:
                # validate default config "file" structure:
                self._validate(data=new_data)

            # validate default config items for the newly added default config:
            config_options.validate(
                data=def_config["config"],
                logger=self.logger,
                raise_with_metadata=False,
            )

        except (ConfigFileValidationError, ConfigValidationError) as err:
            raise ConfigDefaultValidationError(err) from None

        self.__data_rt = new_data_rt
        self.__data = new_data
        self.__contents = self._dump(new_data_rt)

        return name

    @staticmethod
    def get_config_file_path(directory: Path) -> Path:
        """
        Get the path to the configuration file.
        """
        # Try both ".yml" and ".yaml" extensions:
        path_yaml = directory.joinpath("config.yaml")
        if path_yaml.is_file():
            return path_yaml
        path_yml = directory.joinpath("config.yml")
        if path_yml.is_file():
            return path_yml
        return path_yaml

    def _load_file_data(self, config_options: ConfigOptions):
        """Load data from the configuration file (config.yaml or config.yml)."""

        self.__path = self.get_config_file_path(self.directory)
        if not self.path.is_file():
            self.logger.info(
                "No config.yaml found in the configuration directory. Generating "
                "a config.yaml file."
            )
            self.add_default_config(name="default", config_options=config_options)

        yaml = YAML(typ="safe")
        yaml_rt = YAML(typ="rt")
        with self.path.open() as handle:
            contents = handle.read()
            handle.seek(0)
            data = yaml.load(handle)
            handle.seek(0)
            data_rt = yaml_rt.load(handle)

        self.__contents = contents
        self.__data = data
        self.__data_rt = data_rt

    def get_config_item(
        self, config_key: str, name: str, *, raise_on_missing=False, default_value=None
    ) -> Any | None:
        """
        Get a configuration item.

        Parameters
        ----------
        config_key: str
            The name of the configuration within the configuration file.
        name: str
            The name of the configuration item.
        raise_on_missing: bool
            Whether to raise an error if the config item is absent.
        default_value:
            The default value to use when the config item is absent
            (and ``raise_on_missing`` is not specified).
        """
        cfg = self.get_invoc_data(config_key)["config"]
        if raise_on_missing and name not in cfg:
            raise ValueError(f"missing from file: {name!r}")
        return cfg.get(name, default_value)

    def is_item_set(self, config_key: str, name: str) -> bool:
        """
        Determine if a configuration item is set.

        Parameters
        ----------
        config_key: str
            The name of the configuration within the configuration file.
        name: str
            The name of the configuration item.
        """
        try:
            self.get_config_item(config_key, name, raise_on_missing=True)
            return True
        except ValueError:
            return False

    def rename_config_key(self, config_key: str, new_config_key: str) -> None:
        """
        Change the config key of the loaded config.

        Parameters
        ----------
        config_key: str
            The old name of the configuration within the configuration file.
        new_config_key: str
            The new name of the configuration.
        """

        new_data = copy.deepcopy(self.data)
        new_data_rt = copy.deepcopy(self.data_rt)

        new_data["configs"][new_config_key] = new_data["configs"].pop(config_key)
        new_data_rt["configs"][new_config_key] = new_data_rt["configs"].pop(config_key)

        for config in self._configs:
            if config._config_key == config_key:
                config._meta_data["config_key"] = new_config_key
                config._config_key = new_config_key

        self.__data_rt = new_data_rt
        self.__data = new_data
        self.__contents = self._dump(new_data_rt)

    def update_invocation(
        self,
        config_key: str,
        environment_setup: str | None = None,
        match: dict[str, str | list[str]] | None = None,
    ) -> None:
        """
        Modify the invocation parameters of the loaded config.

        Parameters
        ----------
        config_key:
            The name of the configuration within the configuration file.
        environment_setup:
            The new value of the ``environment_setup`` key.
        match:
            The new values to merge into the ``match`` key.
        """

        new_data = copy.deepcopy(self.data)
        new_data_rt = copy.deepcopy(self.data_rt)

        for dat in (new_data, new_data_rt):
            invoc = dat["configs"][config_key]["invocation"]
            if environment_setup:
                invoc["environment_setup"] = environment_setup
            if match:
                invoc["match"].update(match)

        self.__data_rt = new_data_rt
        self.__data = new_data
        self.__contents = self._dump(new_data_rt)



"""
File-system watcher classes.
"""

from __future__ import annotations
from collections.abc import Callable
from datetime import timedelta
from logging import Logger
from pathlib import Path
from typing import cast
from watchdog.observers.polling import PollingObserver
from watchdog.events import (
    FileSystemEvent,
    FileSystemEventHandler,
    PatternMatchingEventHandler,
)


class _PMEHDelegate(PatternMatchingEventHandler):
    def __init__(self, pattern: str, on_modified: Callable[[FileSystemEvent], None]):
        super().__init__(patterns=[pattern])
        self.__on_modified = on_modified

    def on_modified(self, event: FileSystemEvent) -> None:
        self.__on_modified(event)


class _FSEHDelegate(FileSystemEventHandler):
    def __init__(self, on_modified: Callable[[FileSystemEvent], None]) -> None:
        self.__on_modified = on_modified

    def on_modified(self, event: FileSystemEvent) -> None:
        self.__on_modified(event)


class MonitorController:
    """
    Controller for tracking watch files.
    """

    def __init__(
        self,
        workflow_dirs_file_path: str | Path,
        watch_interval: float | timedelta,
        logger: Logger,
    ):
        if isinstance(watch_interval, timedelta):
            self.watch_interval = int(watch_interval.total_seconds())
        else:
            self.watch_interval = int(watch_interval)
        self.workflow_dirs_file_path = Path(workflow_dirs_file_path).absolute()
        self.logger = logger

        if not self.workflow_dirs_file_path.exists():
            self.logger.info(
                f"Watch file does not exist; creating {str(self.workflow_dirs_file_path)}."
            )
            with self.workflow_dirs_file_path.open("wt") as fp:
                fp.write("\n")

        self.logger.info(f"Watching file: {str(self.workflow_dirs_file_path)}")

        self.event_handler = _PMEHDelegate("watch_workflows.txt", self.on_modified)

        self.observer = PollingObserver(timeout=self.watch_interval)
        self.observer.schedule(
            self.event_handler,
            path=cast("str", self.workflow_dirs_file_path.parent),
            recursive=False,
        )

        self.observer.start()

        workflow_paths = self.parse_watch_workflows_file(
            self.workflow_dirs_file_path, logger=self.logger
        )
        self.workflow_monitor = WorkflowMonitor(
            workflow_paths,
            watch_interval=self.watch_interval,
            logger=self.logger,
        )

    @staticmethod
    def parse_watch_workflows_file(
        path: str | Path, logger: Logger
    ) -> list[dict[str, Path]]:
        """
        Parse the file describing what workflows to watch.
        """
        # TODO: and parse element IDs as well; and record which are set/unset.
        with Path(path).open("rt") as fp:
            lns = fp.readlines()

        wks: list[dict[str, Path]] = []
        for ln in lns:
            ln_s = ln.strip()
            if not ln_s:
                continue
            wk_path = Path(ln_s).absolute()
            if not wk_path.is_dir():
                logger.warning(f"{str(wk_path)} is not a workflow")
                continue

            wks.append(
                {
                    "path": wk_path,
                }
            )

        return wks

    def on_modified(self, event: FileSystemEvent):
        """
        Callback when files are modified.
        """
        self.logger.info(f"Watch file modified: {event.src_path!r}")
        wks = self.parse_watch_workflows_file(
            cast("str", event.src_path), logger=self.logger
        )
        self.workflow_monitor.update_workflow_paths(wks)

    def join(self) -> None:
        """
        Join the worker thread.
        """
        self.observer.join()

    def stop(self) -> None:
        """
        Stop this monitor.
        """
        self.observer.stop()
        self.observer.join()  # wait for it to stop!
        self.workflow_monitor.stop()


class WorkflowMonitor:
    """
    Workflow monitor.
    """

    def __init__(
        self,
        workflow_paths: list[dict[str, Path]],
        watch_interval: float | timedelta,
        logger: Logger,
    ):
        if isinstance(watch_interval, timedelta):
            self.watch_interval = int(watch_interval.total_seconds())
        else:
            self.watch_interval = int(watch_interval)

        self.event_handler = _FSEHDelegate(self.on_modified)
        self.workflow_paths = workflow_paths
        self.logger = logger

        self._monitor_workflow_paths()

    def _monitor_workflow_paths(self) -> None:
        observer = PollingObserver(timeout=self.watch_interval)
        self.observer: PollingObserver | None = observer
        for i in self.workflow_paths:
            observer.schedule(
                self.event_handler, path=cast("str", i["path"]), recursive=False
            )
            self.logger.info(f"Watching workflow: {i['path'].name}")

        observer.start()

    def on_modified(self, event: FileSystemEvent):
        """
        Triggered on a workflow being modified.
        """
        self.logger.info(f"Workflow modified: {event.src_path!r}")

    def update_workflow_paths(self, new_paths: list[dict[str, Path]]):
        """
        Change the set of paths to monitored workflows.
        """
        self.logger.info("Updating watched workflows.")
        self.stop()
        self.workflow_paths = new_paths
        self._monitor_workflow_paths()

    def stop(self) -> None:
        """
        Stop this monitor.
        """
        if self.observer:
            self.observer.stop()
            self.observer.join()  # wait for it to stop!
            self.observer = None



"""
Submission enumeration types.
"""

from __future__ import annotations
from dataclasses import dataclass
from enum import Enum

from hpcflow.sdk.core.enums import _ReportableStateData, _ReportableStateEnum


class JobscriptElementState(_ReportableStateEnum):
    """Enumeration to convey a particular jobscript element state as reported by the
    scheduler."""

    #: Waiting for resource allocation.
    pending = _ReportableStateData(
        0,
        "○",
        "yellow",
        "Waiting for resource allocation.",
    )
    #: Waiting for one or more dependencies to finish.
    waiting = _ReportableStateData(
        1,
        "◊",
        "grey46",
        "Waiting for one or more dependencies to finish.",
    )
    #: Executing now.
    running = _ReportableStateData(
        2,
        "●",
        "dodger_blue1",
        "Executing now.",
    )
    #: Previously submitted but is no longer active.
    finished = _ReportableStateData(
        3,
        "■",
        "grey46",
        "Previously submitted but is no longer active.",
    )
    #: Cancelled by the user.
    cancelled = _ReportableStateData(
        4,
        "C",
        "red3",
        "Cancelled by the user.",
    )
    #: The scheduler reports an error state.
    errored = _ReportableStateData(
        5,
        "E",
        "red3",
        "The scheduler reports an error state.",
    )


class SubmissionStatus(Enum):
    """
    The overall status of a submission.
    """

    #: Not yet submitted.
    PENDING = 0
    #: All jobscripts submitted successfully.
    SUBMITTED = 1
    #: Some jobscripts submitted successfully.
    PARTIALLY_SUBMITTED = 2



from __future__ import annotations
from typing import TYPE_CHECKING
from typing_extensions import TypedDict
import pytest

from hpcflow.app import app as hf
from hpcflow.sdk.core.errors import (
    EnvironmentPresetUnknownEnvironmentError,
    InvalidIdentifier,
)
from hpcflow.sdk.core.test_utils import make_actions, make_parameters

if TYPE_CHECKING:
    from hpcflow.sdk.core.actions import Action, ActionEnvironment
    from hpcflow.sdk.core.task import TaskObjective


@pytest.fixture
def act_env_1() -> ActionEnvironment:
    return hf.ActionEnvironment("env_1")


@pytest.fixture
def action_a1(act_env_1: ActionEnvironment) -> Action:
    return hf.Action(commands=[hf.Command("ls")], environments=[act_env_1])


class SchemaKwargs(TypedDict):
    objective: TaskObjective
    actions: list[Action]


@pytest.fixture
def schema_s1_kwargs(action_a1: Action) -> SchemaKwargs:
    return {"objective": hf.TaskObjective("t1"), "actions": [action_a1]}


def test_task_schema_equality() -> None:
    t1a = hf.TaskSchema("t1", actions=[])
    t1b = hf.TaskSchema("t1", actions=[])
    assert t1a == t1b


def test_init_with_str_objective(action_a1: Action) -> None:
    obj_str = "t1"
    obj = hf.TaskObjective(obj_str)
    assert hf.TaskSchema(obj_str, actions=[action_a1]) == hf.TaskSchema(
        obj, actions=[action_a1]
    )


def test_init_with_method_with_underscore(schema_s1_kwargs) -> None:
    hf.TaskSchema(method="my_method", **schema_s1_kwargs)


def test_raise_on_invalid_method_digit(schema_s1_kwargs) -> None:
    with pytest.raises(InvalidIdentifier):
        hf.TaskSchema(method="9", **schema_s1_kwargs)


def test_raise_on_invalid_method_space(schema_s1_kwargs) -> None:
    with pytest.raises(InvalidIdentifier):
        hf.TaskSchema(method="my method", **schema_s1_kwargs)


def test_raise_on_invalid_method_non_alpha_numeric(schema_s1_kwargs) -> None:
    with pytest.raises(InvalidIdentifier):
        hf.TaskSchema(method="_mymethod", **schema_s1_kwargs)


def test_schema_action_validate() -> None:
    p1, p2, p3, p4 = make_parameters(4)
    act_1, act_2, act_3 = make_actions([("p1", "p5"), (("p2", "p5"), "p3"), ("p3", "p4")])
    hf.TaskSchema("t1", actions=[act_1, act_2, act_3], inputs=[p1, p2], outputs=[p3, p4])


def test_schema_action_validate_raise_on_extra_schema_input() -> None:
    # assert raise ValueError
    p1, p2, p3, p4 = make_parameters(4)
    p7 = hf.Parameter("p7")
    act_1, act_2, act_3 = make_actions([("p1", "p5"), (("p2", "p5"), "p3"), ("p3", "p4")])
    with pytest.raises(ValueError):
        hf.TaskSchema(
            "t1", actions=[act_1, act_2, act_3], inputs=[p1, p2, p7], outputs=[p3, p4]
        )


def test_schema_action_validate_raise_on_extra_schema_output() -> None:
    p7 = hf.Parameter("p7")
    p1, p2, p3, p4 = make_parameters(4)
    act_1, act_2, act_3 = make_actions([("p1", "p5"), (("p2", "p5"), "p3"), ("p3", "p4")])
    with pytest.raises(ValueError):
        hf.TaskSchema(
            "t1", actions=[act_1, act_2, act_3], inputs=[p1, p2], outputs=[p3, p4, p7]
        )


def test_schema_action_validate_raise_on_extra_action_input() -> None:
    p1, p2, p3, p4 = make_parameters(4)
    act_1, act_2, act_3 = make_actions(
        [(("p1", "p7"), "p5"), (("p2", "p5"), "p3"), ("p3", "p4")]
    )
    with pytest.raises(ValueError):
        hf.TaskSchema(
            "t1", actions=[act_1, act_2, act_3], inputs=[p1, p2], outputs=[p3, p4]
        )


def test_schema_action_validate_raise_on_extra_action_output() -> None:
    p1, p2, p3, p4 = make_parameters(4)
    act_1, act_2, act_3 = make_actions(
        [("p1", "p5"), (("p2", "p5"), "p3"), ("p3", "p4", "p7")]
    )
    with pytest.raises(ValueError):
        hf.TaskSchema(
            "t1", actions=[act_1, act_2, act_3], inputs=[p1, p2], outputs=[p3, p4]
        )


def test_dot_access_object_list_raise_on_bad_access_attr_name() -> None:
    """Check we can't name a DotAccessObjectList item with a name that collides with a
    method name."""
    ts = hf.TaskSchema("add_object", actions=[])
    with pytest.raises(ValueError):
        hf.TaskSchemasList([ts])


def test_env_preset() -> None:
    p1, p2 = make_parameters(2)
    (act_1,) = make_actions([("p1", "p2")], env="env1")
    hf.TaskSchema(
        "t1",
        inputs=[p1],
        outputs=[p2],
        actions=[act_1],
        environment_presets={"my_preset": {"env1": {"version": 1}}},
    )


def test_env_preset_raise_bad_env() -> None:
    p1, p2 = make_parameters(2)
    (act_1,) = make_actions([("p1", "p2")], env="env1")
    with pytest.raises(EnvironmentPresetUnknownEnvironmentError):
        hf.TaskSchema(
            "t1",
            inputs=[p1],
            outputs=[p2],
            actions=[act_1],
            environment_presets={"my_preset": {"env2": {"version": 1}}},
        )


def test_env_preset_raise_bad_env_no_actions() -> None:
    with pytest.raises(EnvironmentPresetUnknownEnvironmentError):
        hf.TaskSchema(
            "t1",
            environment_presets={"my_preset": {"env1": {"version": 1}}},
        )



from __future__ import annotations

import os

import pytest

from pathlib import Path
from hpcflow.app import app as hf
from hpcflow.sdk.core.actions import SkipReason
from hpcflow.sdk.core.test_utils import make_workflow_to_run_command


def test_compose_commands_no_shell_var(null_config, tmp_path: Path):
    ts = hf.TaskSchema(
        objective="test_compose_commands",
        actions=[hf.Action(commands=[hf.Command(command="Start-Sleep 10")])],
    )
    wk = hf.Workflow.from_template_data(
        template_name="test_compose_commands",
        path=tmp_path,
        tasks=[hf.Task(schema=ts)],
    )
    sub = wk.add_submission()
    assert sub is not None
    js = sub.jobscripts[0]
    run = wk.tasks[0].elements[0].iterations[0].action_runs[0]
    _, shell_vars = run.compose_commands(environments=sub.environments, shell=js.shell)
    assert shell_vars == {0: []}


def test_compose_commands_single_shell_var(null_config, tmp_path: Path):
    ts = hf.TaskSchema(
        objective="test_compose_commands",
        inputs=[hf.SchemaInput("p1")],
        outputs=[hf.SchemaOutput("p1")],
        actions=[
            hf.Action(
                commands=[
                    hf.Command(
                        command="Write-Output (<<parameter:p1>> + 100)",
                        stdout="<<int(parameter:p1)>>",
                    ),
                ],
            ),
        ],
    )
    wk = hf.Workflow.from_template_data(
        template_name="test_compose_commands",
        path=tmp_path,
        tasks=[hf.Task(schema=ts, inputs={"p1": 101})],
    )
    sub = wk.add_submission()
    assert sub is not None
    js = sub.jobscripts[0]
    run = wk.tasks[0].elements[0].iterations[0].action_runs[0]
    _, shell_vars = run.compose_commands(environments=sub.environments, shell=js.shell)
    assert shell_vars == {0: [("outputs.p1", "parameter_p1", "stdout")]}


def test_compose_commands_multi_single_shell_var(null_config, tmp_path: Path):
    ts = hf.TaskSchema(
        objective="test_compose_commands",
        inputs=[hf.SchemaInput("p1")],
        outputs=[hf.SchemaOutput("p1")],
        actions=[
            hf.Action(
                commands=[
                    hf.Command(command="Start-Sleep 10"),
                    hf.Command(
                        command="Write-Output (<<parameter:p1>> + 100)",
                        stdout="<<int(parameter:p1)>>",
                    ),
                ],
            ),
        ],
    )
    wk = hf.Workflow.from_template_data(
        template_name="test_compose_commands",
        path=tmp_path,
        tasks=[hf.Task(schema=ts, inputs={"p1": 101})],
    )
    sub = wk.add_submission()
    assert sub is not None
    js = sub.jobscripts[0]
    run = wk.tasks[0].elements[0].iterations[0].action_runs[0]
    _, shell_vars = run.compose_commands(environments=sub.environments, shell=js.shell)
    assert shell_vars == {0: [], 1: [("outputs.p1", "parameter_p1", "stdout")]}


@pytest.mark.integration
def test_run_dir_diff_new_file(null_config, tmp_path):
    if os.name == "nt":
        command = "New-Item -Path 'new_file.txt' -ItemType File"
    else:
        command = "touch new_file.txt"
    wk = make_workflow_to_run_command(
        command=command,
        requires_dir=True,
        path=tmp_path,
        name="w2",
        overwrite=True,
    )
    wk.submit(wait=True, add_to_known=False, status=False)
    assert wk.get_all_EARs()[0].dir_diff.files_created == ["new_file.txt"]


@pytest.mark.integration
def test_run_skip_reason_upstream_failure(null_config, tmp_path):
    ts = hf.TaskSchema(
        objective="t1",
        inputs=[hf.SchemaInput("p1")],
        outputs=[hf.SchemaInput("p2")],
        actions=[
            hf.Action(
                commands=[
                    hf.Command(
                        command="echo $(( <<parameter:p1>> + 100 ))",
                        stdout="<<parameter:p2>>",
                    ),
                    hf.Command(command="exit 1"),
                ]
            ),
            hf.Action(
                commands=[
                    hf.Command(
                        command="echo $(( <<parameter:p2>> + 100 ))",
                        stdout="<<parameter:p2>>",
                    ),
                ]
            ),  # should be skipped due to failure of action 0
        ],
    )
    wk = hf.Workflow.from_template_data(
        template_name="test_skip_reason",
        path=tmp_path,
        tasks=[hf.Task(schema=ts, inputs={"p1": 100})],
    )
    wk.submit(wait=True, add_to_known=False, status=False)
    runs = wk.get_all_EARs()
    assert not runs[0].success
    assert not runs[1].success
    assert runs[0].skip_reason is SkipReason.NOT_SKIPPED
    assert runs[1].skip_reason is SkipReason.UPSTREAM_FAILURE


@pytest.mark.integration
def test_run_skip_reason_loop_termination(null_config, tmp_path):
    ts = hf.TaskSchema(
        objective="t1",
        inputs=[hf.SchemaInput("p1")],
        outputs=[hf.SchemaInput("p1")],
        actions=[
            hf.Action(
                commands=[
                    hf.Command(
                        command="echo $(( <<parameter:p1>> + 100 ))",
                        stdout="<<int(parameter:p1)>>",
                    ),
                ]
            ),
        ],
    )
    loop_term = hf.Rule(path="outputs.p1", condition={"value.equal_to": 300})
    wk = hf.Workflow.from_template_data(
        template_name="test_skip_reason",
        path=tmp_path,
        tasks=[hf.Task(schema=ts, inputs={"p1": 100})],
        loops=[
            hf.Loop(name="my_loop", tasks=[0], termination=loop_term, num_iterations=3)
        ],
    )
    # loop should terminate after the second iteration
    wk.submit(wait=True, add_to_known=False, status=False)
    runs = wk.get_all_EARs()

    assert runs[0].get("outputs.p1") == 200
    assert runs[1].get("outputs.p1") == 300
    assert not runs[2].get("outputs.p1")

    assert runs[0].success
    assert runs[1].success
    assert not runs[2].success

    assert runs[0].skip_reason is SkipReason.NOT_SKIPPED
    assert runs[1].skip_reason is SkipReason.NOT_SKIPPED
    assert runs[2].skip_reason is SkipReason.LOOP_TERMINATION



import os
import sys
from pathlib import Path
import pytest

from hpcflow.app import app as hf
from hpcflow.sdk.core import SKIPPED_EXIT_CODE
from hpcflow.sdk.core.skip_reason import SkipReason


@pytest.mark.integration
@pytest.mark.parametrize("exit_code", [0, 1, 98, -1, -123124])
def test_action_exit_code_parsing(null_config, tmp_path: Path, exit_code: int):
    act = hf.Action(commands=[hf.Command(command=f"exit {exit_code}")])
    s1 = hf.TaskSchema(
        objective="t1",
        actions=[act],
    )
    t1 = hf.Task(schema=[s1])
    wk = hf.Workflow.from_template_data(tasks=[t1], template_name="test", path=tmp_path)
    wk.submit(wait=True, add_to_known=False)
    recorded_exit = wk.get_EARs_from_IDs([0])[0].exit_code
    if os.name == "posix":
        # exit code from bash wraps around:
        exit_code %= 256
    assert recorded_exit == exit_code


@pytest.mark.integration
def test_bad_action_py_script_exit_code(null_config, tmp_path):
    s1 = hf.TaskSchema(
        objective="t1",
        actions=[
            hf.Action(
                script="<<script:bad_script.py>>",  # raises SyntaxError
                script_exe="python_script",
                environments=[hf.ActionEnvironment(environment="python_env")],
            )
        ],
    )
    t1 = hf.Task(schema=[s1])
    wk = hf.Workflow.from_template_data(
        tasks=[t1], template_name="bad_script_test", path=tmp_path
    )
    wk.submit(wait=True, add_to_known=False)
    recorded_exit = wk.get_EARs_from_IDs([0])[0].exit_code
    assert recorded_exit == 1


@pytest.mark.integration
@pytest.mark.parametrize("exit_code", [0, 1, 98, -1, -123124])
def test_action_py_script_specified_exit_code(null_config, tmp_path, exit_code):
    s1 = hf.TaskSchema(
        objective="t1",
        inputs=[hf.SchemaInput("exit_code")],
        actions=[
            hf.Action(
                script="<<script:script_exit_test.py>>",
                script_exe="python_script",
                script_data_in="direct",
                environments=[hf.ActionEnvironment(environment="python_env")],
            )
        ],
    )
    t1 = hf.Task(schema=[s1], inputs={"exit_code": exit_code})
    wk = hf.Workflow.from_template_data(
        tasks=[t1], template_name="script_exit_test", path=tmp_path
    )
    wk.submit(wait=True, add_to_known=False)
    recorded_exit = wk.get_EARs_from_IDs([0])[0].exit_code
    if os.name == "posix":
        # exit code from bash wraps around:
        exit_code %= 256
    assert recorded_exit == exit_code


@pytest.mark.integration
def test_skipped_action_same_element(null_config, tmp_path):
    s1 = hf.TaskSchema(
        objective="t1",
        inputs=[hf.SchemaInput("p1")],
        outputs=[hf.SchemaOutput("p2"), hf.SchemaOutput("p3")],
        actions=[
            hf.Action(
                commands=[
                    hf.Command(
                        command=f"echo <<parameter:p1>>", stdout="<<parameter:p2>>"
                    ),
                    hf.Command(command=f"exit 1"),
                ],
            ),
            hf.Action(  # should be skipped
                commands=[
                    hf.Command(
                        command=f"echo <<parameter:p2>>", stdout="<<parameter:p3>>"
                    ),
                    hf.Command(command=f"exit 0"),  # exit code should be ignored
                ],
            ),
        ],
    )
    t1 = hf.Task(schema=s1, inputs={"p1": 101})
    wk = hf.Workflow.from_template_data(
        tasks=[t1], template_name="test_skip", path=tmp_path
    )
    wk.submit(wait=True, add_to_known=False, status=False)

    runs = wk.get_EARs_from_IDs([0, 1])
    exit_codes = [i.exit_code for i in runs]
    is_skipped = [i.skip for i in runs]

    assert exit_codes == [1, SKIPPED_EXIT_CODE]
    assert is_skipped == [0, 1]


@pytest.mark.integration
def test_two_skipped_actions_same_element(null_config, tmp_path):
    s1 = hf.TaskSchema(
        objective="t1",
        inputs=[hf.SchemaInput("p1")],
        outputs=[hf.SchemaOutput("p2"), hf.SchemaOutput("p3"), hf.SchemaOutput("p4")],
        actions=[
            hf.Action(
                commands=[
                    hf.Command(
                        command=f"echo <<parameter:p1>>", stdout="<<parameter:p2>>"
                    ),
                    hf.Command(command=f"exit 1"),
                ],
            ),
            hf.Action(  # should be skipped
                commands=[
                    hf.Command(
                        command=f"echo <<parameter:p2>>", stdout="<<parameter:p3>>"
                    ),
                    hf.Command(command=f"exit 0"),  # exit code should be ignored
                ],
            ),
            hf.Action(  # should be skipped
                commands=[
                    hf.Command(
                        command=f"echo <<parameter:p3>>", stdout="<<parameter:p4>>"
                    ),
                    hf.Command(command=f"exit 0"),  # exit code should be ignored
                ],
            ),
        ],
    )
    t1 = hf.Task(schema=s1, inputs={"p1": 101})
    wk = hf.Workflow.from_template_data(
        tasks=[t1], template_name="test_skip_two_actions", path=tmp_path
    )
    wk.submit(wait=True, add_to_known=False, status=False)

    runs = wk.get_EARs_from_IDs([0, 1, 2])
    exit_codes = [i.exit_code for i in runs]
    skip_reasons = [i.skip_reason for i in runs]

    assert exit_codes == [1, SKIPPED_EXIT_CODE, SKIPPED_EXIT_CODE]
    assert skip_reasons == [
        SkipReason.NOT_SKIPPED,
        SkipReason.UPSTREAM_FAILURE,
        SkipReason.UPSTREAM_FAILURE,
    ]


@pytest.mark.integration
@pytest.mark.skipif(
    condition=sys.platform == "win32",
    reason="`combine_jobscript_std` not implemented on Windows.",
)
def test_combine_jobscript_std_true(null_config, tmp_path):
    out_msg = "hello stdout!"
    err_msg = "hello stderr!"
    s1 = hf.TaskSchema(
        objective="t1",
        actions=[
            hf.Action(
                commands=[
                    hf.Command(command=f'echo "{out_msg}"'),
                    hf.Command(command=f'>&2 echo "{err_msg}"'),
                ],
            )
        ],
    )
    t1 = hf.Task(schema=s1)
    wk = hf.Workflow.from_template_data(
        tasks=[t1],
        template_name="test_combine_jobscript_std",
        path=tmp_path,
        resources={"any": {"combine_jobscript_std": True}},
    )
    wk.submit(wait=True, add_to_known=False, status=False)

    jobscript = wk.submissions[0].jobscripts[0]

    assert jobscript.resources.combine_jobscript_std

    out_err_path = jobscript.get_std_out_err_path()
    out_path = jobscript._get_stdout_path()
    err_path = jobscript._get_stderr_path()

    assert out_err_path.is_file()
    assert not out_path.is_file()
    assert not err_path.is_file()

    assert out_err_path.read_text().strip() == f"{out_msg}\n{err_msg}"


@pytest.mark.integration
def test_combine_jobscript_std_false(null_config, tmp_path):
    out_msg = "hello stdout!"
    err_msg = "hello stderr!"
    s1 = hf.TaskSchema(
        objective="t1",
        actions=[
            hf.Action(
                commands=[
                    hf.Command(command=f'echo "{out_msg}"'),
                    hf.Command(command=f'>&2 echo "{err_msg}"'),
                ],
                rules=[
                    hf.ActionRule(
                        rule=hf.Rule(
                            path="resources.os_name",
                            condition={"value.equal_to": "posix"},
                        )
                    )
                ],
            ),
            hf.Action(
                commands=[
                    hf.Command(command=f'Write-Output "{out_msg}"'),
                    hf.Command(command=f'$host.ui.WriteErrorLine("{err_msg}")'),
                ],
                rules=[
                    hf.ActionRule(
                        rule=hf.Rule(
                            path="resources.os_name",
                            condition={"value.equal_to": "nt"},
                        )
                    )
                ],
            ),
        ],
    )
    t1 = hf.Task(schema=s1)
    wk = hf.Workflow.from_template_data(
        tasks=[t1],
        template_name="test_combine_jobscript_std",
        path=tmp_path,
        resources={"any": {"combine_jobscript_std": False}},
    )
    wk.submit(wait=True, add_to_known=False, status=False)

    jobscript = wk.submissions[0].jobscripts[0]

    assert not jobscript.resources.combine_jobscript_std

    out_err_path = jobscript.direct_std_out_err_path
    out_path = jobscript.direct_stdout_path
    err_path = jobscript.direct_stderr_path

    assert not out_err_path.is_file()
    assert out_path.is_file()
    assert err_path.is_file()

    assert out_path.read_text().strip() == out_msg
    assert err_path.read_text().strip() == err_msg


@pytest.mark.integration
def test_write_app_logs_true(null_config, tmp_path):

    p1_vals = [101, 102]
    t1 = hf.Task(
        schema=hf.task_schemas.test_t1_conditional_OS,
        sequences=[hf.ValueSequence("inputs.p1", values=p1_vals)],
    )
    wk = hf.Workflow.from_template_data(
        tasks=[t1],
        template_name="test_write_app_logs",
        path=tmp_path,
        config={
            "log_file_level": "debug"
        },  # ensure there is something to write to the log
        resources={"any": {"write_app_logs": True}},
    )
    wk.submit(wait=True, add_to_known=False, status=False)

    run_0 = wk.tasks[0].elements[0].action_runs[0]
    run_1 = wk.tasks[0].elements[1].action_runs[0]

    run_0_log_path = run_0.get_app_log_path()
    run_1_log_path = run_1.get_app_log_path()

    assert run_0_log_path.is_file()
    assert run_1_log_path.is_file()


@pytest.mark.integration
def test_write_app_logs_false(null_config, tmp_path):

    p1_vals = [101, 102]
    t1 = hf.Task(
        schema=hf.task_schemas.test_t1_conditional_OS,
        sequences=[hf.ValueSequence("inputs.p1", values=p1_vals)],
    )
    wk = hf.Workflow.from_template_data(
        tasks=[t1],
        template_name="test_write_app_logs",
        path=tmp_path,
        config={
            "log_file_level": "debug"
        },  # ensure there is something to write to the log
        resources={"any": {"write_app_logs": False}},
    )
    wk.submit(wait=True, add_to_known=False, status=False)

    run_0 = wk.tasks[0].elements[0].action_runs[0]
    run_1 = wk.tasks[0].elements[1].action_runs[0]

    run_0_log_path = run_0.get_app_log_path()
    run_1_log_path = run_1.get_app_log_path()

    assert not wk.submissions[0].app_log_path.is_dir()
    assert not run_0_log_path.is_file()
    assert not run_1_log_path.is_file()


@pytest.mark.integration
def test_jobscript_start_end_times_equal_to_first_and_last_run_start_end_times(
    null_config, tmp_path
):

    t1 = hf.Task(
        schema=hf.task_schemas.test_t1_conditional_OS,
        sequences=[hf.ValueSequence(path="inputs.p1", values=list(range(2)))],
    )
    wk = hf.Workflow.from_template_data(
        template_name="test_jobscript_start_end_times",
        path=tmp_path,
        tasks=[t1],
    )
    wk.submit(wait=True, add_to_known=False, status=False)

    js = wk.submissions[0].jobscripts[0]
    runs = wk.get_all_EARs()
    assert len(runs) == 2

    # jobsript has two runs, so start time should be start time of first run:
    assert js.start_time == runs[0].start_time

    # ...and end time should be end time of second run:
    assert js.end_time == runs[1].end_time



import pytest
from hpcflow.sdk.core.utils import get_file_context
from hpcflow.app import app as hf


@pytest.mark.integration
def test_api_make_and_submit_workflow(null_config, tmp_path):
    with get_file_context("hpcflow.tests.data", "workflow_1.yaml") as file_path:
        wk = hf.make_and_submit_workflow(
            file_path,
            path=tmp_path,
            status=False,
            add_to_known=False,
            wait=True,
        )
        p2 = wk.tasks[0].elements[0].outputs.p2
        assert isinstance(p2, hf.ElementParameter)
        assert p2.value == "201"


@pytest.mark.integration
def test_api_make_and_submit_demo_workflow(null_config, tmp_path):
    wk = hf.make_and_submit_demo_workflow(
        "workflow_1",
        path=tmp_path,
        status=False,
        add_to_known=False,
        wait=True,
    )
    p2 = wk.tasks[0].elements[0].outputs.p2
    assert isinstance(p2, hf.ElementParameter)
    assert p2.value == "201"



"""
CLI components for demonstration code.
"""

from __future__ import annotations
from pathlib import Path
from random import randint
from typing import TYPE_CHECKING
import click

from hpcflow.sdk.core.workflow import Workflow
from hpcflow.sdk.core.utils import get_process_stamp
from hpcflow.sdk.cli_common import (
    format_option,
    path_option,
    name_option,
    overwrite_option,
    store_option,
    ts_fmt_option,
    ts_name_fmt_option,
    variables_option,
    js_parallelism_option,
    wait_option,
    add_to_known_opt,
    print_idx_opt,
    tasks_opt,
    cancel_opt,
    submit_status_opt,
    make_status_opt,
    add_sub_opt,
)
from hpcflow.sdk.submission.submission import Submission

if TYPE_CHECKING:
    from collections.abc import Iterable
    from typing import Literal
    from ..app import BaseApp


def get_demo_software_CLI(app: BaseApp):
    """Generate the CLI to provide an example software."""

    @click.group()
    def demo_software():
        pass

    @demo_software.command("doSomething")
    @click.option("--infile1", "-i1", type=click.Path(exists=True), required=True)
    @click.option("--infile2", "-i2", type=click.Path(exists=True), required=True)
    @click.option("--value", "-v")
    @click.option("--out", "-o")
    def demo_do_something(
        infile1: Path, infile2: Path, value: str | None = None, out: str | None = None
    ):
        click.echo("trying to do something")

        with Path(infile1).open("r") as handle:
            file_id_1 = int(handle.readline().strip())
        with Path(infile2).open("r") as handle:
            file_id_2 = int(handle.readline().strip())

        if out is None:
            out = "outfile.txt"
        out_path = Path(out)
        with out_path.open("a") as handle:
            handle.write("{}\n".format(randint(0, int(1e6))))
            handle.write(
                "{} Generated by `doSomething --infile1 {} --infile2 {}`.\n".format(
                    get_process_stamp(), infile1, infile2
                )
            )
            if value:
                handle.write("{} Value: {}\n".format(get_process_stamp(), value))
            handle.write(
                "{} Original file ID: {}: {}\n".format(
                    get_process_stamp(), infile1, file_id_1
                )
            )
            handle.write(
                "{} Original file ID: {}: {}\n".format(
                    get_process_stamp(), infile2, file_id_2
                )
            )

    return demo_software


def get_demo_workflow_CLI(app: BaseApp):
    """Generate the CLI to provide access to builtin demo workflows."""

    def list_callback(ctx: click.Context, param, value: bool):
        if not value or ctx.resilient_parsing:
            return
        # TODO: format with Rich with a one-line description
        click.echo("\n".join(app.list_demo_workflows()))
        ctx.exit()

    @click.group()
    @click.option(
        "-l",
        "--list",
        help="Print available builtin demo workflows.",
        is_flag=True,
        is_eager=True,
        expose_value=False,
        callback=list_callback,
    )
    def demo_workflow():
        """Interact with builtin demo workflows."""
        pass

    @demo_workflow.command("make")
    @click.argument("workflow_name")
    @format_option
    @path_option
    @name_option
    @overwrite_option
    @store_option
    @ts_fmt_option
    @ts_name_fmt_option
    @variables_option
    @make_status_opt
    @add_sub_opt
    def make_demo_workflow(
        workflow_name: str,
        format: Literal["json", "yaml"] | None,
        path: Path | None,
        name: str | None,
        overwrite: bool,
        store: str,
        ts_fmt: str | None = None,
        ts_name_fmt: str | None = None,
        variables: Iterable[tuple[str, str]] = (),
        status: bool = True,
        add_submission: bool = False,
    ):
        wk_or_sub = app.make_demo_workflow(
            workflow_name=workflow_name,
            template_format=format,
            path=path,
            name=name,
            overwrite=overwrite,
            store=store,
            ts_fmt=ts_fmt,
            ts_name_fmt=ts_name_fmt,
            variables=dict(variables),
            status=status,
            add_submission=add_submission,
        )
        if add_submission:
            assert isinstance(wk_or_sub, Submission)
            click.echo(wk_or_sub.workflow.path)
        else:
            assert isinstance(wk_or_sub, Workflow)
            click.echo(wk_or_sub.path)

    @demo_workflow.command("go")
    @click.argument("workflow_name")
    @format_option
    @path_option
    @name_option
    @overwrite_option
    @store_option
    @ts_fmt_option
    @ts_name_fmt_option
    @variables_option
    @js_parallelism_option
    @wait_option
    @add_to_known_opt
    @print_idx_opt
    @tasks_opt
    @cancel_opt
    @submit_status_opt
    def make_and_submit_demo_workflow(
        workflow_name: str,
        format: Literal["json", "yaml"] | None,
        path: Path | None,
        name: str | None,
        overwrite: bool,
        store: str,
        ts_fmt: str | None = None,
        ts_name_fmt: str | None = None,
        variables: Iterable[tuple[str, str]] = (),
        js_parallelism: bool | None = None,
        wait: bool = False,
        add_to_known: bool = True,
        print_idx: bool = False,
        tasks: list[int] | None = None,
        cancel: bool = False,
        status: bool = True,
    ):
        out = app.make_and_submit_demo_workflow(
            workflow_name=workflow_name,
            template_format=format,
            path=path,
            name=name,
            overwrite=overwrite,
            store=store,
            ts_fmt=ts_fmt,
            ts_name_fmt=ts_name_fmt,
            variables=dict(variables),
            JS_parallelism=js_parallelism,
            wait=wait,
            add_to_known=add_to_known,
            return_idx=print_idx,
            tasks=tasks,
            cancel=cancel,
            status=status,
        )
        if print_idx:
            assert isinstance(out, tuple)
            click.echo(out[1])

    @demo_workflow.command("copy")
    @click.argument("workflow_name")
    @click.argument("destination")
    @click.option("--doc/--no-doc", default=True)
    def copy_demo_workflow(workflow_name: str, destination: str, doc: bool):
        app.copy_demo_workflow(name=workflow_name, dst=destination, doc=doc)

    @demo_workflow.command("show")
    @click.argument("workflow_name")
    @click.option("--syntax/--no-syntax", default=True)
    @click.option("--doc/--no-doc", default=True)
    def show_demo_workflow(workflow_name: str, syntax: bool, doc: bool):
        app.show_demo_workflow(workflow_name, syntax=syntax, doc=doc)

    return demo_workflow



"""
A direct job "scheduler" that just runs immediate subprocesses.
"""

from __future__ import annotations
import shutil
import signal
from typing import overload, cast, TYPE_CHECKING
from typing_extensions import override, TypeAlias
import psutil

from hpcflow.sdk.typing import hydrate
from hpcflow.sdk.submission.enums import JobscriptElementState
from hpcflow.sdk.submission.schedulers import Scheduler

if TYPE_CHECKING:
    from collections.abc import Callable, Mapping, Sequence
    from typing import Any, ClassVar
    from ...config.types import SchedulerConfigDescriptor
    from ..jobscript import Jobscript
    from ..shells.base import Shell

DirectRef: TypeAlias = "tuple[int, list[str]]"


def _is_process_cmdline_equal(proc: psutil.Process, cmdline: list[str]) -> bool:
    """Check if the `cmdline` of a psutil `Process` is equal to the specified
    `cmdline`."""
    try:
        if proc.cmdline() == cmdline:
            return True
        else:
            return False
    except (psutil.NoSuchProcess, psutil.ZombieProcess):
        # process no longer exists or, on unix, process has completed but still has a
        # record
        return False


class DirectScheduler(Scheduler[DirectRef]):
    """
    A direct scheduler, that just runs jobs immediately as direct subprocesses.

    The correct subclass (:py:class:`DirectPosix` or :py:class:`DirectWindows`) should
    be used to create actual instances.

    """

    @classmethod
    @override
    def process_resources(
        cls, resources, scheduler_config: SchedulerConfigDescriptor
    ) -> None:
        """Perform scheduler-specific processing to the element resources.

        Note
        ----
        This mutates `resources`.
        """
        return

    @override
    def get_submit_command(
        self,
        shell: Shell,
        js_path: str,
        deps: dict[Any, tuple[Any, ...]],
    ) -> list[str]:
        """
        Get the concrete submission command.
        """
        return shell.get_direct_submit_command(js_path)

    @staticmethod
    def __kill_processes(
        procs: list[psutil.Process],
        sig: signal.Signals = signal.SIGTERM,
        timeout: float | None = None,
        on_terminate: Callable[[psutil.Process], object] | None = None,
    ):
        all_procs: list[psutil.Process] = []
        for process in procs:
            all_procs.append(process)
            all_procs.extend(process.children(recursive=True))

        for process in all_procs:
            try:
                process.send_signal(sig)
            except psutil.NoSuchProcess:
                pass
        _, alive = psutil.wait_procs(all_procs, timeout=timeout, callback=on_terminate)
        for process in alive:
            process.kill()

    @staticmethod
    def __get_jobscript_processes(js_refs: list[DirectRef]) -> list[psutil.Process]:
        procs: list[psutil.Process] = []
        for p_id, p_cmdline in js_refs:
            try:
                proc_i = psutil.Process(p_id)
            except psutil.NoSuchProcess:
                # process might have completed already
                continue
            if _is_process_cmdline_equal(proc_i, p_cmdline):
                procs.append(proc_i)
        return procs

    @overload
    @override
    @classmethod
    def wait_for_jobscripts(cls, js_refs: list[DirectRef]) -> None: ...

    @overload
    @classmethod
    def wait_for_jobscripts(
        cls,
        js_refs: list[DirectRef],
        *,
        callback: Callable[[psutil.Process], None],
    ) -> list[psutil.Process]: ...

    @classmethod
    def wait_for_jobscripts(
        cls,
        js_refs: list[DirectRef],
        *,
        callback: Callable[[psutil.Process], None] | None = None,
    ) -> list[psutil.Process] | None:
        """Wait until the specified jobscripts have completed."""
        procs = cls.__get_jobscript_processes(js_refs)
        (gone, alive) = psutil.wait_procs(procs, callback=callback)
        assert not alive
        return gone if callback else None

    @override
    def get_job_state_info(
        self, *, js_refs: Sequence[DirectRef] | None = None
    ) -> Mapping[str, JobscriptElementState]:
        """Query the scheduler to get the states of all of this user's jobs, optionally
        filtering by specified job IDs.

        Jobs that are not in the scheduler's status output will not appear in the output
        of this method."""
        info: dict[str, JobscriptElementState] = {}
        for p_id, p_cmdline in js_refs or ():
            if self.is_jobscript_active(p_id, p_cmdline):
                # as far as the "scheduler" is concerned, all elements are running:
                info[str(p_id)] = JobscriptElementState.running

        return info

    @override
    def cancel_jobs(
        self,
        js_refs: list[DirectRef],
        jobscripts: list[Jobscript] | None = None,
    ):
        """
        Cancel some jobs.
        """

        js_proc_id: dict[int, Jobscript]

        def callback(proc: psutil.Process):
            try:
                js_proc_id[proc.pid]
            except KeyError:
                # child process of one of the jobscripts
                self._app.submission_logger.debug(
                    f"jobscript child process ({proc.pid}) killed"
                )
                return

        procs = self.__get_jobscript_processes(js_refs)
        self._app.submission_logger.info(
            f"cancelling {self.__class__.__name__} jobscript processes: {procs}."
        )
        js_proc_id = {i.pid: jobscripts[idx] for idx, i in enumerate(procs) if jobscripts}
        self.__kill_processes(procs, timeout=3, on_terminate=callback)
        print(f"Cancelled {len(procs)} jobscript{'s' if len(procs) > 1 else ''}.")
        self._app.submission_logger.info("jobscripts cancel command executed.")

    def is_jobscript_active(self, process_ID: int, process_cmdline: list[str]):
        """Query if a jobscript is running.

        Note that a "running" jobscript might be waiting on upstream jobscripts to
        complete.

        """
        try:
            proc = psutil.Process(process_ID)
        except psutil.NoSuchProcess:
            return False
        return _is_process_cmdline_equal(proc, process_cmdline)

    def get_std_out_err_filename(self, js_idx: int, **kwargs) -> str:
        """File name of combined standard output and error streams."""
        return f"js_{js_idx}_std.log"

    def get_stdout_filename(self, js_idx: int, **kwargs) -> str:
        """File name of the standard output stream file."""
        return f"js_{js_idx}_stdout.log"

    def get_stderr_filename(self, js_idx: int, **kwargs) -> str:
        """File name of the standard error stream file."""
        return f"js_{js_idx}_stderr.log"


@hydrate
class DirectPosix(DirectScheduler):
    """
    A direct scheduler for POSIX systems.

    """


@hydrate
class DirectWindows(DirectScheduler):
    """
    A direct scheduler for Windows.

    """

    @override
    def get_submit_command(
        self, shell: Shell, js_path: str, deps: dict[Any, tuple[Any, ...]]
    ) -> list[str]:
        cmd = super().get_submit_command(shell, js_path, deps)
        # `Start-Process` (see `Jobscript._launch_direct_js_win`) seems to resolve the
        # executable, which means the process's `cmdline` might look different to what we
        # record; so let's resolve it ourselves:
        cmd[0] = cast("str", shutil.which(cmd[0]))
        return cmd



"""
Utilities for discovering what persistence store implementation to use.
"""

from __future__ import annotations
from typing import TYPE_CHECKING
from hpcflow.sdk.persistence.json import JSONPersistentStore
from hpcflow.sdk.persistence.zarr import ZarrPersistentStore, ZarrZipPersistentStore

if TYPE_CHECKING:
    from collections.abc import Mapping
    from .base import PersistentStore

# Because of python/mypy#4717, we need to disable an error here:
# mypy: disable-error-code="type-abstract"
_ALL_STORE_CLS: Mapping[str, type[PersistentStore]] = {
    "zarr": ZarrPersistentStore,
    "zip": ZarrZipPersistentStore,
    "json": JSONPersistentStore,
    # "json-single": JSONPersistentStore,  # TODO
}
# Without that, there's literally no way to write the above with a sane type.

#: The persistence formats supported.
ALL_STORE_FORMATS = tuple(_ALL_STORE_CLS)
#: The persistence formats supported for creation.
ALL_CREATE_STORE_FORMATS = tuple(
    k for k, v in _ALL_STORE_CLS.items() if v._features.create
)


def store_cls_from_str(store_format: str) -> type[PersistentStore]:
    """
    Get the class that implements the persistence store from its name.
    """
    try:
        return _ALL_STORE_CLS[store_format]
    except KeyError:
        raise ValueError(f"Store format {store_format!r} not known.")



from __future__ import annotations

from typing import TYPE_CHECKING, overload

if TYPE_CHECKING:
    from numpy.typing import NDArray


@overload
def get_2D_idx(idx: int, num_cols: int) -> tuple[int, int]: ...


@overload
def get_2D_idx(idx: NDArray, num_cols: int) -> tuple[NDArray, NDArray]: ...


def get_2D_idx(idx: int | NDArray, num_cols: int) -> tuple[int | NDArray, int | NDArray]:
    """Convert a 1D index to a 2D index, assuming items are arranged in a row-major
    order."""
    row_idx = idx // num_cols
    col_idx = idx % num_cols
    return (row_idx, col_idx)


def get_1D_idx(
    row_idx: int | NDArray, col_idx: int | NDArray, num_cols: int
) -> int | NDArray:
    """Convert a 2D (row, col) index into a 1D index, assuming items are arranged in a
    row-major order."""
    return row_idx * num_cols + col_idx


def split_arr(arr: NDArray, metadata_size: int) -> list[tuple[NDArray, NDArray]]:
    """Split a 1D integer array into a list of tuples, each containing a metadata array
    and a data array, where the size of each (metadata + data) sub-array is specified as
    the integer immediately before each (metadata + data) sub-array.

    Parameters
    ----------
    arr
        One dimensional integer array to split.
    metadata_size
        How many elements to include in the metadata array. This can be zero.

    Returns
    -------
    sub_arrs
        List of tuples of integer arrays. The integers that define the sizes of the
        sub-arrays are excluded.

    Examples
    --------
    >>> split_arr(np.array([4, 0, 1, 2, 3, 4, 1, 4, 5, 6]), metadata_size=1)
    [(array([0]), array([1, 2, 3])), (array([1]), array([4, 5, 6]))]

    """
    count = 0
    block_start = 0
    sub_arrs = []
    while count < len(arr):
        size = arr[block_start]
        start = block_start + 1
        end = start + size
        metadata_i = arr[start : start + metadata_size]
        sub_arr_i = arr[start + metadata_size : end]
        sub_arrs.append((metadata_i, sub_arr_i))
        count += size + 1
        block_start = end
    return sub_arrs



import pytest
from hpcflow.app import app as hf


def test_SGE_process_resources_multi_core_with_parallel_env(null_config):

    scheduler_config = {
        "parallel_environments": {
            None: {"num_cores": [1, 1, 1]},  # [start, step, stop]
            "my_parallel_env": {"num_cores": [2, 1, 32]},
        }
    }

    scheduler = hf.SGEPosix()
    resources = hf.ElementResources(num_cores=2, SGE_parallel_env="my_parallel_env")

    scheduler.process_resources(resources, scheduler_config)

    assert resources.num_cores == 2
    assert resources.SGE_parallel_env == "my_parallel_env"


def test_SGE_process_resources_raises_on_single_core_with_parallel_env(null_config):

    scheduler_config = {
        "parallel_environments": {
            None: {"num_cores": [1, 1, 1]},  # [start, step, stop]
            "my_parallel_env": {"num_cores": [2, 1, 32]},
        }
    }

    scheduler = hf.SGEPosix()
    resources = hf.ElementResources(num_cores=1, SGE_parallel_env="my_parallel_env")

    with pytest.raises(ValueError):
        scheduler.process_resources(resources, scheduler_config)



"""
Information about the Python runtime.
"""

from __future__ import annotations
from importlib import import_module
from logging import Logger
import os
import platform
import re
import socket
import sys
from pathlib import Path
from typing import Any, ClassVar

from rich.table import Table
from rich.console import Console


class RunTimeInfo:
    """Get useful run-time information, including the executable name used to
    invoke the CLI, in the case a PyInstaller-built executable was used.

    Parameters
    ----------
    name:
        Application name.
    package_name:
        Application package name.
    version:
        Application version.
    logger:
        Where to write logging versions.
    """

    def __init__(
        self, name: str, package_name: str, version: str, logger: Logger
    ) -> None:
        is_frozen: bool = getattr(sys, "frozen", False) and hasattr(sys, "_MEIPASS")
        bundle_dir = (
            sys._MEIPASS
            if is_frozen and hasattr(sys, "_MEIPASS")
            else os.path.dirname(os.path.abspath(__file__))
        )

        #: Application name.
        self.name = name.split(".")[0]  # if name is given as __name__ # TODO: what?
        #: Application package name.
        self.package_name = package_name
        #: Application version.
        self.version = version
        #: Whether this is a frozen application.
        self.is_frozen = is_frozen
        #: Working directory.
        self.working_dir = os.getcwd()
        #: Where to write log messages.
        self.logger = logger
        #: Host that this is running on.
        self.hostname = socket.gethostname()

        #: Whether this application is inside iPython.
        self.in_ipython = False
        #: Whether this application is being used interactively.
        self.is_interactive = False
        #: Whether this application is being used in test mode.
        self.in_pytest = False  # set in `conftest.py`
        #: Whether this application is being run from the CLI.
        self.from_CLI = False  # set in CLI

        if self.is_frozen:
            #: The bundle directory, if frozen.
            self.bundle_dir = Path(bundle_dir)
        else:
            #: The path to Python itself.
            self.python_executable_path = Path(sys.executable)

            try:
                get_ipython  # type: ignore
            except NameError:
                pass
            else:
                self.in_ipython = True

            if hasattr(sys, "ps1"):
                self.is_interactive = True

        #: The Python version.
        self.python_version = platform.python_version()
        #: Whether the application is in a virtual environment.
        self.is_venv = hasattr(sys, "real_prefix") or sys.base_prefix != sys.prefix
        #: Whether the application is in a Conda virtual environment.
        self.is_conda_venv = "CONDA_PREFIX" in os.environ

        #: From `sys.prefix`. If running in a virtual environment, this will point to the
        #: environment directory. If not running in a virtual environment, this will
        #: point to the Python installation root.
        self.sys_prefix = getattr(sys, "prefix", None)
        #: From `sys.base_prefix`. This will be equal to `sys_prefix` (`sys.prefix`) if
        #: not running within a virtual environment. However, if running within a virtual
        #: environment, this will be the Python installation directory, and `sys_prefix`
        #: will be equal to the virtual environment directory.
        self.sys_base_prefix = getattr(sys, "base_prefix", None)
        #: The old base prefix, from `sys.real_prefix`. Compatibility version of
        #: :py:attr:`sys_base_prefix`.
        self.sys_real_prefix = getattr(sys, "real_prefix", None)
        #: The Conda prefix, if defined.
        self.conda_prefix = os.environ.get("CONDA_PREFIX")

        try:
            #: The virtual environment path.
            self.venv_path: str | list[str] | None = self.__set_venv_path()
        except ValueError:
            self.venv_path = None

        self.logger.debug(
            f"is_frozen: {self.is_frozen!r}"
            f"{f' ({self.executable_name!r})' if self.is_frozen else ''}"
        )
        self.logger.debug(
            f"is_venv: {self.is_venv!r}"
            f"{f' ({self.sys_prefix!r})' if self.is_venv else ''}"
        )
        self.logger.debug(
            f"is_conda_venv: {self.is_conda_venv!r}"
            f"{f' ({self.conda_prefix!r})' if self.is_conda_venv else ''}"
        )
        # TODO: investigate
        # if self.is_venv and self.is_conda_venv:
        #     msg = (
        #         "Running in a nested virtual environment (conda and non-conda). "
        #         "Environments may not be re-activate in the same order in associated, "
        #         "subsequent invocations of hpcflow."
        #     )
        #     warnings.warn(msg)

    def to_dict(self) -> dict[str, Any]:
        """
        Serialize this class as a dictionary.
        """
        out = {
            "name": self.name,
            "package_name": self.package_name,
            "version": self.version,
            "is_frozen": self.is_frozen,
            "working_dir": self.working_dir,
            "logger": self.logger,
            "hostname": self.hostname,
            "python_version": self.python_version,
            "invocation_command": self.invocation_command,
            "in_ipython": self.in_ipython,
            "in_pytest": self.in_pytest,
            "from_CLI": self.from_CLI,
        }
        if self.is_frozen:
            out.update(
                {
                    "executable_name": self.executable_name,
                    "resolved_executable_name": self.resolved_executable_name,
                    "executable_path": self.executable_path,
                    "resolved_executable_path": self.resolved_executable_path,
                }
            )
        else:
            out.update(
                {
                    "is_interactive": self.is_interactive,
                    "script_path": self.script_path,
                    "resolved_script_path": self.resolved_script_path,
                    "python_executable_path": self.python_executable_path,
                    "is_venv": self.is_venv,
                    "is_conda_venv": self.is_conda_venv,
                    "sys_prefix": self.sys_prefix,
                    "sys_base_prefix": self.sys_base_prefix,
                    "sys_real_prefix": self.sys_real_prefix,
                    "conda_prefix": self.conda_prefix,
                    "venv_path": self.venv_path,
                }
            )
        return out

    def __repr__(self) -> str:
        out = f"{self.__class__.__name__}("
        out += ", ".join(f"{k}={v!r}" for k, v in self.to_dict().items())
        return out

    def __set_venv_path(self) -> str | list[str]:
        out: list[str] = []
        if self.sys_prefix is not None:
            out.append(self.sys_prefix)
        elif self.conda_prefix is not None:
            out.append(self.conda_prefix)
        if not out:
            raise ValueError("Not running in a virtual environment!")
        if len(out) == 1:
            return out[0]
        else:
            return out

    def get_activate_env_command(self):
        """
        Get the command to activate the virtual environment.
        """
        pass

    def get_deactivate_env_command(self):
        """
        Get the command to deactivate the virtual environment.
        """
        pass

    def show(self) -> None:
        """
        Display the information known by this class as a human-readable table.
        """
        tab = Table(show_header=False, box=None)
        tab.add_column()
        tab.add_column()
        for k, v in self.to_dict().items():
            tab.add_row(k, str(v))

        console = Console()
        console.print(tab)

    @property
    def executable_path(self) -> Path | None:
        """Get the path that the user invoked to launch the frozen app, if the app is
        frozen.

        If the user launches the app via a symbolic link, then this returns that link,
        whereas `executable_path_resolved` returns the actual frozen app path.

        """
        return Path(sys.argv[0]) if self.is_frozen else None

    @property
    def resolved_executable_path(self) -> Path | None:
        """Get the resolved path to the frozen app that the user launched, if the app is
        frozen.

        In a one-file app, this is the path to the bootloader. In the one-folder app, this
        is the path to the executable.

        References
        ----------
        [1] https://pyinstaller.org/en/stable/runtime-information.html#using-sys-executable-and-sys-argv-0

        """
        return Path(sys.executable) if self.is_frozen else None

    @property
    def executable_name(self) -> str | None:
        """Get the name of the frozen app executable, if the app is frozen.

        If the user launches the app via a symbolic link, then this returns the name of
        that link, whereas `resolved_executable_name` returns the actual frozen app file
        name.

        """
        return None if (p := self.executable_path) is None else p.name

    @property
    def resolved_executable_name(self) -> str | None:
        """Get the resolved name of the frozen app executable, if the app is frozen."""
        return None if (p := self.resolved_executable_path) is None else p.name

    @property
    def script_path(self) -> Path | None:
        """Get the path to the Python script used to invoked this instance of the app, if
        the app is not frozen."""
        return None if self.is_frozen else Path(sys.argv[0])

    @property
    def resolved_script_path(self) -> Path | None:
        """Get the resolved path to the Python script used to invoked this instance of the
        app, if the app is not frozen."""
        return None if (p := self.script_path) is None else p.resolve()

    # For removing a trailing '.cmd' from a filename
    __CMD_TRIM: ClassVar[re.Pattern[str]] = re.compile(r"\.cmd$")

    @property
    def invocation_command(self) -> tuple[str, ...]:
        """Get the command that was used to invoke this instance of the app."""
        if self.is_frozen:
            # (this also works if we are running tests using the frozen app)
            return (str(self.resolved_executable_path),)
        elif self.from_CLI:
            script = str(self.resolved_script_path)
            if os.name == "nt":
                # cannot reproduce locally, but on Windows GHA runners, if pytest is
                # invoked via `hpcflow test`, `resolved_script_path` seems to be the
                # batch script wrapper (ending in .cmd) rather than the Python entry point
                # itself, so trim if off:
                script = self.__CMD_TRIM.sub("", script)  # Work with 3.8 too
                # script = script.removesuffix(".cmd")
            return (str(self.python_executable_path), script)
        else:
            app_module = import_module(self.package_name)
            CLI_path = Path(*app_module.__path__, "cli.py")
            return (str(self.python_executable_path), str(CLI_path))



import os
import time

import pytest

from hpcflow.app import app as hf


@pytest.mark.integration
def test_non_snippet_script_execution(null_config, tmp_path):
    test_str = "non-snippet script!"
    script_name = "my_script.py"
    script_contents = f'print("{test_str}")'

    if os.name == "nt":
        cmd = f"Set-Content -Path {script_name} -Value '{script_contents}'"
    else:
        cmd = f"echo '{script_contents}' > {script_name}"

    act_1 = hf.Action(
        commands=[hf.Command(cmd)],
    )
    act_2 = hf.Action(
        script="my_script.py",
        script_exe="python_script",
        script_data_in="direct",
        environments=[hf.ActionEnvironment(environment="python_env")],
    )
    s1 = hf.TaskSchema(
        objective="t1",
        inputs=[hf.SchemaInput(parameter=hf.Parameter("p1"))],
        actions=[act_1, act_2],
    )

    t1 = hf.Task(schema=s1, inputs={"p1": 101})
    wk = hf.Workflow.from_template_data(
        tasks=[t1],
        template_name="non_snippet_script_test",
        workflow_name="non_snippet_script_test",
        overwrite=True,
        path=tmp_path,
    )
    wk.submit(wait=True, add_to_known=False, status=False)

    std_out = wk.submissions[0].jobscripts[0].direct_stdout_path.read_text().strip()
    assert std_out.endswith(test_str)



"""
Miscellaneous persistence-related helpers.
"""

from __future__ import annotations
from getpass import getpass
from typing import TYPE_CHECKING

from hpcflow.sdk.core.errors import WorkflowNotFoundError

if TYPE_CHECKING:
    from typing import Callable, TypeVar
    from fsspec import AbstractFileSystem  # type: ignore

    T = TypeVar("T")


def ask_pw_on_auth_exc(
    f: Callable[..., T], *args, add_pw_to: str | None = None, **kwargs
) -> tuple[T, str | None]:
    """
    Run the given function on the given arguments and add a password if the function
    fails with an SSHException.
    """
    from paramiko.ssh_exception import SSHException

    try:
        out = f(*args, **kwargs)
        pw = None

    except SSHException:
        pw = getpass()

        if not add_pw_to:
            kwargs["password"] = pw
        else:
            kwargs[add_pw_to] = {**kwargs[add_pw_to], "password": pw}

        out = f(*args, **kwargs)

    return out, pw


def infer_store(path: str, fs: AbstractFileSystem) -> str:
    """Identify the store type using the path and file system parsed by fsspec.

    Parameters
    ----------
    fs
        fsspec file system

    """

    # TODO: raise WorkflowNotFoundError if the path does not exist
    # TODO: raise MalformedWorkflowError if a known store type cannot be inferred

    # try to identify store type just from the path string:
    if path.endswith(".zip"):
        store_fmt = "zip"

    elif path.endswith(".json"):
        store_fmt = "json-single"

    else:
        # look at the directory contents:
        if fs.glob(f"{path}/.zattrs"):
            store_fmt = "zarr"
        elif fs.glob(f"{path}/metadata.json"):
            store_fmt = "json"
        else:
            raise WorkflowNotFoundError(path, fs)

    return store_fmt
